
# 正規表現でtextファイルと同じやり方で読み込むのですべて手打ちする．

__version__      = '0.3.7'
__license__      = 'Apache License 2.0'
__author__       = 'yu9824'
__copyright__    = 'Copyright © 2021 yu9824'
__author_email__ = 'yu.9824.job@gmail.com'
__user_id__ = 'yu9824'
__app_name__ = 'yikit'
__url__ = 'http://github.com/yu9824/yikit/'

# __all__ = [
#     'models',
#     'tools',
#     'feature_selection',
#     ]